/*
**      $VER: StartUp.c 1.0
**
**      Library startup-code and function table definition
**
**      Based on CExecLib sources from Andreas R. Kleinert
*/

#define __USE_SYSBASE

#include <exec/types.h>
#include <exec/memory.h>
#include <exec/libraries.h>
#include <exec/execbase.h>
#include <exec/resident.h>
#include <exec/initializers.h>
#include <proto/exec.h>

#include "ixdtbase.h"
#include "Funcs.h"

extern ULONG __saveds __stdargs L_OpenLibs(void);
extern void  __saveds __stdargs L_CloseLibs(void);
struct IXDTBase * __saveds __asm InitLib( register __a6 struct ExecBase      *sysbase,
                    register __a0 struct SegList       *seglist,
                    register __d0 struct IXDTBase *ixdtb);
struct IXDTBase * __saveds __asm OpenLib( register __a6 struct IXDTBase *IXDTBase);
struct SegList * __saveds __asm CloseLib( register __a6 struct IXDTBase *IXDTBase);
struct SegList * __saveds __asm ExpungeLib( register __a6 struct IXDTBase *ixdtb);
ULONG __saveds __asm ExtFuncLib(void);
LONG __saveds __asm LibStart(void)
{
 return(-1);
}

extern APTR FuncTab [];
extern struct MyDataInit DataTab;

struct InitTable                       /* do not change */
{
 ULONG              LibBaseSize;
 APTR              *FunctionTable;
 struct MyDataInit *DataTable;
 APTR               InitLibTable;
} InitTab =
{
 sizeof(struct IXDTBase),
 &FuncTab[0],
 &DataTab,
 InitLib
};

APTR FuncTab [] =                  /* iX Datatype required functions */
{
 OpenLib,
 CloseLib,
 ExpungeLib,
 ExtFuncLib,

 DTF_Identify,
 DTF_FileStruct,
 DTF_ProcessChunk,

 (APTR) ((LONG)-1)
};

extern struct IXDTBase *IXDTBase;

struct IXDTBase * __saveds __asm InitLib( register __a6 struct ExecBase      *sysbase,
                    register __a0 struct SegList       *seglist,
                    register __d0 struct IXDTBase *ixdtb)
{
 IXDTBase = ixdtb;

 IXDTBase->dtb_SysBase = sysbase;
 IXDTBase->dtb_SegList = seglist;

 if(L_OpenLibs()) return(IXDTBase);

 L_CloseLibs();

 return(NULL);
}

struct IXDTBase * __saveds __asm OpenLib( register __a6 struct IXDTBase *IXDTBase)
{
 IXDTBase->dtb_LibNode.lib_OpenCnt++;
 IXDTBase->dtb_LibNode.lib_Flags &= ~LIBF_DELEXP;
 return(IXDTBase);
}

struct SegList * __saveds __asm CloseLib( register __a6 struct IXDTBase *IXDTBase)
{
 IXDTBase->dtb_LibNode.lib_OpenCnt--;

 if(!IXDTBase->dtb_LibNode.lib_OpenCnt)
  {
   if(IXDTBase->dtb_LibNode.lib_Flags & LIBF_DELEXP)
    {
     return( ExpungeLib(IXDTBase) );
    }
  }

 return(NULL);
}

struct SegList * __saveds __asm ExpungeLib( register __a6 struct IXDTBase *ixdtb)
{
 struct IXDTBase *IXDTBase = ixdtb;
 struct SegList       *seglist;

 if(!IXDTBase->dtb_LibNode.lib_OpenCnt)
  {
   ULONG negsize, possize, fullsize;
   UBYTE *negptr = (UBYTE *) IXDTBase;

   seglist = IXDTBase->dtb_SegList;

   Remove((struct Node *)IXDTBase);

   L_CloseLibs();

   negsize  = IXDTBase->dtb_LibNode.lib_NegSize;
   possize  = IXDTBase->dtb_LibNode.lib_PosSize;
   fullsize = negsize + possize;
   negptr  -= negsize;

   FreeMem(negptr, fullsize);

   return(seglist);
  }
  
 IXDTBase->dtb_LibNode.lib_Flags |= LIBF_DELEXP;

 return(NULL);
}

ULONG __saveds __asm ExtFuncLib(void)
{
 return(NULL);
}

struct IXDTBase *IXDTBase = NULL;

ULONG XCEXIT       = NULL;  /* these symbols may be referenced by    */
ULONG _XCEXIT      = NULL;  /* some functions of sc.lib, but should  */
ULONG ONBREAK      = NULL;  /* never be used inside a shared library */
ULONG _ONBREAK     = NULL;
ULONG base         = NULL;
ULONG _base        = NULL;
ULONG ProgramName  = NULL;
ULONG _ProgramName = NULL;
ULONG StackPtr     = NULL;
ULONG _StackPtr    = NULL;
ULONG oserr        = NULL;
ULONG _oserr       = NULL;
ULONG OSERR        = NULL;
ULONG _OSERR       = NULL;

void __regargs __chkabort(void) { }  /* a shared library cannot be    */
void __regargs _CXBRK(void)     { }  /* CTRL-C aborted when doing I/O */

